/*
                                  resample.c

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*
*       Part of:        Swarp
*
*       Author:         E.BERTIN (IAP)
*
*       Contents:       Resampling procedures
*
*       Last modify:    14/06/2002
*
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#ifdef HAVE_CONFIG_H
#include	"config.h"
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "define.h"
#include "types.h"
#include "globals.h"
#include "fits/fitscat.h"
#include "fitswcs.h"
#include "data.h"
#include "field.h"
#include "header.h"
#include "interpolate.h"
#include "prefs.h"
#include "resample.h"
#ifdef USE_THREADS
#include "threads.h"
#endif
#include "weight.h"
#include "wcs/wcs.h"

#ifdef USE_THREADS
pthread_t	thread[THREADS_NMAX];
pthread_mutex_t	startmutex, readymutex;
pthread_cond_t	startcond, readycond;
int		startsem[THREADS_NMAX], stopsem[THREADS_NMAX];
#endif

fieldstruct	*infield,*inwfield, *field, *wfield;
ikernelstruct	*ikernel[THREADS_NMAX];
wcsstruct	*wcsinp[THREADS_NMAX], *wcsoutp[THREADS_NMAX];
double		rawposp[THREADS_NMAX][NAXIS], rawmin[NAXIS], stepover[NAXIS],
		*wcsbuf[THREADS_NMAX],*rawbuf[THREADS_NMAX],
		ascale;
PIXTYPE		*outbuf[THREADS_NMAX],*outwbuf[THREADS_NMAX];
int		proc[THREADS_NMAX],
		*oversamp,
		noversamp, oversampflag, width, naxis, readysem,
		nproc, nprocmax;
char		padbuf[FBSIZE];

/*------------------------------ function -----------------------------------*/
void		*pthread_warp_lines(void *arg);


/****** resample_field *******************************************************
PROTO	void resample_field(fieldstruct **pinfield, fieldstruct **pinwfield,
			fieldstruct *outfield, fieldstruct *outwfield,
			interpenum *interptype)
PURPOSE	Resample an image.
INPUT	Input pointer to field structure pointer,
	Input pointer weight-map field structure pointer,
	Output total field structure pointer,
	Output total weight-map field structure pointer,
	Interpolation type.
	Output
OUTPUT	-.
NOTES	The structure pointers pointed by pinfield and and pinwfield are
	updated and point to the resampled fields on output.
AUTHOR	E. Bertin (IAP)
VERSION	14/06/2002
 ***/
void	resample_field(fieldstruct **pinfield, fieldstruct **pinwfield,
		fieldstruct *outfield, fieldstruct *outwfield,
		interpenum *interptype)
  {
#ifdef USE_THREADS
   static pthread_attr_t	pthread_attr;
#endif
   wcsstruct		wcsloc,
			*wcs;
   static char		filename[MAXCHAR],filename2[MAXCHAR],
			resampext1[MAXCHAR], resampext2[MAXCHAR];
   char			*pstr;
   static double	rawmax[NAXIS], rawpos[NAXIS];
   double		ascale1;
   int			d,y, height, size, p;

  infield = *pinfield;
  inwfield = *pinwfield;

/* Create new file name */
  strcpy(filename2, infield->rfilename);
  if (pstr=strrchr(filename2, '.'))
    *pstr = '\0';
  if (infield->frameno)
    sprintf(filename, "%s/%s.%04d%s", prefs.resampdir_name, filename2,
				infield->frameno, prefs.resamp_suffix);
  else
    sprintf(filename, "%s/%s%s", prefs.resampdir_name, filename2,
				prefs.resamp_suffix);

/* We make a copy of the output field */
  field = inherit_field(filename, outfield, FIELD_WRITE);
  field->backmean = infield->backmean;
  field->backsig = infield->backsig;
  field->gain = infield->gain;
  field->fieldno = infield->fieldno;
  field->fscale = infield->fscale;

/* Now modify some characteristics of the output file */
/* We use a small "dirty" trick to frame the output */
  wcsloc = *(outfield->wcs);	/* Copy the CONTENT only */
  wcs = infield->wcs;
  naxis = wcsloc.naxis;
  for (d=0; d<naxis;d++)
    {
    wcsloc.outmin[d] = wcs->outmin[d] < 1.0? 1.0:wcs->outmin[d];
    wcsloc.outmax[d] = wcs->outmax[d] > wcsloc.naxisn[d]?
                        (double)wcsloc.naxisn[d] : wcs->outmax[d];
    wcsloc.crpix[d] -= (wcsloc.outmin[d]-1.0);
    wcsloc.naxisn[d] = (wcsloc.outmax[d]-wcsloc.outmin[d]+1);
    if (wcsloc.naxisn[d]<=0)
/*-- Image out of output bounds: create a dummy one with only 1 pixel */
      {
      wcsloc.naxisn[d]=1;
      wcsloc.outmax[d] = wcsloc.outmin[d];
      }
    }

  wcs = field->wcs = copy_wcs(&wcsloc);
  write_wcs(field->tab, wcs);

/* Update field characteristics */
  width = field->width = field->tab->naxisn[0];
  field->height = 1;
  for (d=1; d<naxis; d++)
    field->height *= field->tab->naxisn[d];
  height = field->height;
  field->npix = field->width*field->height;

/* Add relevant information to FITS headers */
  writefitsinfo_field(field);

/* Write image header */
  if (open_cat(field->cat, WRITE_ONLY) != RETURN_OK)
      error(EXIT_FAILURE, "*Error*: cannot open for writing ", filename);
  if (prefs.removetmp_flag && prefs.combine_flag)
    add_cleanupfilename(filename);
  QFWRITE(field->tab->headbuf, field->tab->headnblock*FBSIZE,
	field->cat->file, filename);

/* Get resampled suffix extension if available */
  strcpy(resampext1, prefs.resamp_suffix);
  *resampext2 = '\0';
  if (pstr=strrchr(resampext1, '.'))
    {
    strcpy(resampext2, pstr);
    *pstr = '\0';
    }

/* Now go on with output weight-map */
  if (infield->frameno)
    sprintf(filename, "%s/%s%s.%04d.weight%s", prefs.resampdir_name, filename2,
					resampext1, infield->frameno,
					resampext2);
  else
    sprintf(filename, "%s/%s%s.weight%s", prefs.resampdir_name, filename2,
					resampext1, resampext2);

  wfield = init_weight(filename, field);
  if (inwfield)
    set_weightconv(inwfield);
/* Add relevant information to FITS headers */
  writefitsinfo_field(wfield);
  wfield->cflags = 0;

/* Write image header */
  if (open_cat(wfield->cat, WRITE_ONLY) != RETURN_OK)
    error(EXIT_FAILURE, "*Error*: cannot open for writing ", filename);
  if (prefs.removetmp_flag && prefs.combine_flag)
    add_cleanupfilename(filename);
  QFWRITE(wfield->tab->headbuf, wfield->tab->headnblock*FBSIZE,
	wfield->cat->file, filename);

/* Prepare oversampling stuff */
  ascale = 1.0;
  noversamp = 1;
  oversamp = prefs.oversamp;
  for (d=0; d<naxis; d++)
    {
    ascale1 = wcs->wcsscale[d]/infield->wcs->wcsscale[d];
    ascale *= ascale1;
    if (!oversamp[d])
/*-- Automatic oversampling mode */
      oversamp[d] = ascale1>1.0 ? (int)(ascale1+0.5) : 1;
    noversamp *= oversamp[d];
    }
  oversampflag = (noversamp>1);
  if (ascale < 1.0)
    ascale = 1.0;


#ifdef USE_THREADS
/* Set up multi-threading stuff */
  nprocmax = nproc = prefs.nthreads;
  QPTHREAD_MUTEX_INIT(&startmutex, NULL);
  QPTHREAD_MUTEX_INIT(&readymutex, NULL);
  QPTHREAD_COND_INIT(&startcond, NULL);
  QPTHREAD_COND_INIT(&readycond, NULL);
  QPTHREAD_ATTR_INIT(&pthread_attr);
  QPTHREAD_ATTR_SETDETACHSTATE(&pthread_attr, PTHREAD_CREATE_JOINABLE);
#else
  nprocmax = nproc = 1;
#endif

  for (p=0; p<nprocmax; p++)
    {
/*-- Allocate memory for the output buffers that contain the final data in */
/*-- internal format (PIXTYPE) */
    QMALLOC(outbuf[p], PIXTYPE, width);
    QMALLOC(outwbuf[p], PIXTYPE, width);
/*-- Provide memory space for the current astrometric line */
    QMALLOC(wcsbuf[p], double, naxis*width*noversamp);
    QMALLOC(rawbuf[p], double, naxis*width*noversamp);
/*-- Initialize interpolation kernel */
    ikernel[p] = init_ikernel(interptype, naxis);
/*-- Make copies of the WCS structure (the WCS library is not reentrant) */
    wcsinp[p] = copy_wcs(infield->wcs);
    wcsoutp[p] = copy_wcs(field->wcs);
    proc[p] = p;

#ifdef USE_THREADS
/* Set Start and Stop semaphores to 0 (only useful for multhreading) */
    startsem[p] = 0;
    stopsem[p] = 0;
/*-- Start threads! */
    QPTHREAD_CREATE(&thread[p], &pthread_attr, &pthread_warp_lines, &proc[p]);
#endif
    }

/* Initialize the astrometric vector */
  for (d=0; d<naxis;d++)
    {
    rawmin[d] = rawpos[d] = 0.5 + 0.5*(stepover[d]=1.0/oversamp[d]);
    rawmax[d] = (double)wcs->naxisn[d];
    }

/* Go on with resampling!! */
/* Loop over output lines: this can be over more than 1 (Y) dimension */
  for (y=0; y<height; y+=nproc)
    {
#ifdef USE_THREADS
/*-- Check that there are enough remaining lines per processor */
    if ((height-y)<nprocmax)
      nproc = height-y;
#endif

    if (!(y%(4*nprocmax)))
      NPRINTF(OUTPUT, "\33[1M> Resampling line:%7d / %-7d\n\33[1A",
	y, height);

    for (p=0; p<nproc; p++)
      {
/*-- Set content for (basic) parallelizing */
      for (d=1; d<naxis; d++)
        rawposp[p][d] = rawpos[d];
/*---- Update coordinate vector */
      for (d=1; d<naxis; d++)
        if ((rawpos[d]+=1.0)<= rawmax[d])
          break;
        else
          rawpos[d] = rawmin[d];
      }

#ifdef USE_THREADS
/*-- Set the multiple semaphore */
    QPTHREAD_MUTEX_LOCK(&readymutex);
    readysem = nproc;
    QPTHREAD_MUTEX_UNLOCK(&readymutex);

/*-- Send start signal */
    QPTHREAD_MUTEX_LOCK(&startmutex);
    for (p=nproc; p<nprocmax; p++)
      stopsem[p]=1;
    QPTHREAD_COND_BROADCAST(&startcond);
/*-- Notify that the alarm clock rang! */
    for (p=0; p<nproc; p++)
      startsem[p] = 1;
    QPTHREAD_MUTEX_UNLOCK(&startmutex);
#else
    warp_line(proc[0]);
#endif

#ifdef USE_THREADS
/*-- Wait for all threads to finish before writing */
    QPTHREAD_MUTEX_LOCK(&readymutex);
    for (p=nproc; p--;)
      if (readysem)
        QPTHREAD_COND_WAIT(&readycond, &readymutex);
    QPTHREAD_MUTEX_UNLOCK(&readymutex);
#endif

    for (p=0; p<nproc; p++)
      {
      write_body(field->tab, outbuf[p], width);
      write_body(wfield->tab, outwbuf[p], width);
      }
    }

#ifdef USE_THREADS
/* Terminate all remaining sleeping threads */
  QPTHREAD_MUTEX_LOCK(&startmutex);
/*-- Notify that the alarm clocks rang! */
  for (p=0; p<nproc; p++)
    stopsem[p] = 1;
  QPTHREAD_COND_BROADCAST(&startcond);
/*-- Notify that the alarm clocks rang! */
  for (p=0; p<nproc; p++)
    startsem[p] = 1;
  QPTHREAD_MUTEX_UNLOCK(&startmutex);
/* Clean up multi-threading stuff */
  for (p=0; p<nprocmax; p++)
    QPTHREAD_JOIN(thread[p], NULL);
  QPTHREAD_MUTEX_DESTROY(&startmutex);
  QPTHREAD_MUTEX_DESTROY(&readymutex);
  QPTHREAD_COND_DESTROY(&startcond);
  QPTHREAD_COND_DESTROY(&readycond);
  QPTHREAD_ATTR_DESTROY(&pthread_attr);
#endif

/* FITS padding*/
  size = PADEXTRA(field->tab->tabsize);
  if (size)
    QFWRITE(padbuf, (size_t)size, field->cat->file, field->filename);
  size = PADEXTRA(wfield->tab->tabsize);
  if (size)
    QFWRITE(padbuf, (size_t)size, wfield->cat->file, wfield->filename);

/* Close files */
  close_cat(field->cat);
  close_cat(wfield->cat);

/* Return back new field pointers */
  *pinfield = field;
  *pinwfield = wfield;

/* Free memory */
  for (p=0; p<nprocmax; p++)
    {
    free(outbuf[p]);
    free(outwbuf[p]);
    free(wcsbuf[p]);
    free(rawbuf[p]);
/*-- Free interpolation kernel */
    free_ikernel(ikernel[p]);
    }
  end_field(infield);
  if (inwfield)
    end_field(inwfield);

  return;
  }

#ifdef USE_THREADS


/****** pthread_warp_lines ****************************************************
PROTO	void *pthread_warp_lines(void *arg)
PURPOSE	thread that takes care of resampling image lines
INPUT	Pointer to the thread number.
OUTPUT	-.
NOTES	-.
AUTHOR	E. Bertin (IAP)
VERSION	25/04/2002
 ***/
void	*pthread_warp_lines(void *arg)
  {
   int	p;

  p = *((int *)arg);
  while (1)
    {
/*-- Wait for synchronization signal */
    QPTHREAD_MUTEX_LOCK(&startmutex);
    if (!startsem[p])
      QPTHREAD_COND_WAIT(&startcond, &startmutex);
    if (stopsem[p])
      break;
    startsem[p] = 0;
    QPTHREAD_MUTEX_UNLOCK(&startmutex);
/*-- Go! */
    warp_line(p);
/*-- Send ready signal */
    QPTHREAD_MUTEX_LOCK(&readymutex);
    QPTHREAD_COND_SIGNAL(&readycond);
/*-- Notify that the alarm clock rang! */
    readysem--;
    QPTHREAD_MUTEX_UNLOCK(&readymutex);
    }

  QPTHREAD_MUTEX_UNLOCK(&startmutex);

  pthread_exit(NULL);
  }

/****** cancel_resample_threads ***********************************************
PROTO	void clean_resample_threads(void)
PURPOSE	Cancel remaining active threads
INPUT	-.
OUTPUT	-.
NOTES	-.
AUTHOR	E. Bertin (IAP)
VERSION	25/04/2002
 ***/
void	cancel_resample_threads(void)
  {
   int	p;

  for (p=0; p<nproc; p++)
    QPTHREAD_CANCEL(thread[p]);

  return;
  }

#endif

/****** warp_line *******************************************************
PROTO	void warp_line(int p)
PURPOSE	Resample an image line.
INPUT	Thread number.
OUTPUT	-.
NOTES	-.
AUTHOR	E. Bertin (IAP)
VERSION	22/04/2002
 ***/
void	warp_line(int p)
  {
   wcsstruct		*wcsin,*wcsout;
   double		rawposover[NAXIS],
			*rawpos, *wcspos, *rawbufc,
			dpix,dpixw;
   PIXTYPE		*out, *outw,
			pix,pixw;
   int			nstepover[NAXIS],
			d, o, x, ninput;

  rawpos = rawposp[p];
  wcsin = wcsinp[p];
  wcsout = wcsoutp[p];
/* Compute all alpha's and delta's for the current line */
  wcspos = wcsbuf[p];
  rawpos[0] = rawmin[0];
  if (oversampflag)
    {
    for (d=0; d<naxis; d++)
      nstepover[d] = 0;
    for (x=width; x--; (*rawpos)+=1.0)
      {
      for (d=0; d<naxis; d++)
        rawposover[d] = rawpos[d];
        for (o=noversamp; o--; wcspos+=naxis)
          {
          raw_to_wcs(wcsout, rawposover, wcspos);
/*----- Update coordinate vector */
          for (d=0; d<naxis; d++)
            {
            rawposover[d] += stepover[d];
            if (++nstepover[d]<oversamp[d])
              break;
            else
              {
              rawposover[d] = rawpos[d];
              nstepover[d] = 0;
              }
            }
          }
        }
      }
    else
      for (x=width; x--; wcspos+=naxis, (*rawpos)+=1.0)
        raw_to_wcs(wcsout, rawpos, wcspos);
/*-- Interpolate input images over the next output line */
    wcspos = wcsbuf[p];
    rawbufc = rawbuf[p];
    for (x=width*noversamp; x--; wcspos+=naxis, rawbufc+=naxis)
      if (*wcspos == WCS_NOCOORD)
        *rawbufc = WCS_NOCOORD;
      else
        wcs_to_raw(wcsin, wcspos, rawbufc);
    rawbufc = rawbuf[p];
    out = outbuf[p];
    outw = outwbuf[p];
    if (oversampflag)
      for (x=width; x--;)
        {
        ninput = 0;
        dpix = dpixw = 0.0;
        for (o=noversamp; o--; rawbufc+=naxis)
          if (*rawbufc != WCS_NOCOORD)
            {
            interpolate_pix(infield, inwfield, ikernel[p],rawbufc,&pix,&pixw);
            if (pixw<BIG)
              {
              dpix += (double)pix;
              dpixw += (double)pixw;
              ninput++;
              }
            }
        *(out++) = ninput? dpix/ninput: 0.0;
        *(outw++) = ninput? (ninput/dpixw)*ascale:0.0;
        }
    else
      for (x=width; x--; rawbufc+=naxis)
        if (*rawbufc == WCS_NOCOORD)
          *(out++) = *(outw++) = 0.0;
        else
          {
          interpolate_pix(infield, inwfield, ikernel[p], rawbufc,out++,outw);
/*------- Convert variance to weight */
          *outw = (*outw < BIG) ? 1.0/(*outw) : 0.0;
          outw++;
          }

  return;
  }

